@@ -2,7 +2,7 @@
  *
  * pg_isready --- checks the status of the PostgreSQL server
  *
- * Copyright (c) 2013-2015, PostgreSQL Global Development Group
+ * Copyright (c) 2013-2019, PostgreSQL Global Development Group
  *
  * src/bin/scripts/pg_isready.c
  *
@@ -11,6 +11,7 @@
 
 #include "postgres_fe.h"
 #include "common.h"
+#include "common/logging.h"
 
 #define DEFAULT_CONNECT_TIMEOUT "3"
 
@@ -63,6 +64,7 @@ main(int argc, char **argv)
 		{NULL, 0, NULL, 0}
 	};
 
+	pg_logging_init(argv[0]);
 	progname = get_progname(argv[0]);
 	set_pglocale_pgservice(argv[0], PG_TEXTDOMAIN("pgscripts"));
 	handle_help_version_opts(argc, argv, progname, help);
@@ -102,8 +104,8 @@ main(int argc, char **argv)
 
 	if (optind < argc)
 	{
-		fprintf(stderr, _("%s: too many command-line arguments (first is \"%s\")\n"),
-				progname, argv[optind]);
+		pg_log_error("too many command-line arguments (first is \"%s\")",
+					 argv[optind]);
 		fprintf(stderr, _("Try \"%s --help\" for more information.\n"), progname);
 
 		/*
@@ -139,7 +141,7 @@ main(int argc, char **argv)
 		opts = PQconninfoParse(pgdbname, &errmsg);
 		if (opts == NULL)
 		{
-			fprintf(stderr, _("%s: %s"), progname, errmsg);
+			pg_log_error("%s", errmsg);
 			exit(PQPING_NO_ATTEMPT);
 		}
 	}
@@ -147,7 +149,7 @@ main(int argc, char **argv)
 	defs = PQconndefaults();
 	if (defs == NULL)
 	{
-		fprintf(stderr, _("%s: could not fetch default options\n"), progname);
+		pg_log_error("could not fetch default options");
 		exit(PQPING_NO_ATTEMPT);
 	}
 
@@ -233,5 +235,5 @@ help(const char *progname)
 	printf(_("  -p, --port=PORT          database server port\n"));
 	printf(_("  -t, --timeout=SECS       seconds to wait when attempting connection, 0 disables (default: %s)\n"), DEFAULT_CONNECT_TIMEOUT);
 	printf(_("  -U, --username=USERNAME  user name to connect as\n"));
-	printf(_("\nReport bugs to <pgsql-bugs@postgresql.org>.\n"));
+	printf(_("\nReport bugs to <pgsql-bugs@lists.postgresql.org>.\n"));
 }
